@@ -16,7 +16,6 @@
 """
 OpenStack Swift client library used internally
 """
-
 import socket
 import requests
 import logging
@@ -24,7 +23,8 @@ import warnings
 
 from distutils.version import StrictVersion
 from requests.exceptions import RequestException, SSLError
-from six.moves.urllib.parse import quote as _quote
+from six.moves import http_client
+from six.moves.urllib.parse import quote as _quote, unquote
 from six.moves.urllib.parse import urlparse, urlunparse
 from time import sleep, time
 import six
@@ -34,6 +34,9 @@ from swiftclient.exceptions import ClientException
 from swiftclient.utils import (
     LengthWrapper, ReadableToIterable, parse_api_response)
 
+# Defautl is 100, increase to 256
+http_client._MAXHEADERS = 256
+
 AUTH_VERSIONS_V1 = ('1.0', '1', 1)
 AUTH_VERSIONS_V2 = ('2.0', '2', 2)
 AUTH_VERSIONS_V3 = ('3.0', '3', 3)
@@ -43,6 +46,7 @@ USER_METADATA_TYPE = tuple('x-%s-meta-' % type_ for type_ in
 try:
     from logging import NullHandler
 except ImportError:
+    # Added in Python 2.7
     class NullHandler(logging.Handler):
         def handle(self, record):
             pass
@@ -100,6 +104,36 @@ def http_log(args, kwargs, resp, body):
         log_method("RESP BODY: %s", body)
 
 
+def parse_header_string(data):
+    if six.PY2:
+        if isinstance(data, six.text_type):
+            # Under Python2 requests only returns binary_type, but if we get
+            # some stray text_type input, this should prevent unquote from
+            # interpretting %-encoded data as raw code-points.
+            data = data.encode('utf8')
+        try:
+            unquoted = unquote(data).decode('utf8')
+        except UnicodeDecodeError:
+            try:
+                return data.decode('utf8')
+            except UnicodeDecodeError:
+                return quote(data).decode('utf8')
+    else:
+        if isinstance(data, six.binary_type):
+            # Under Python3 requests only returns text_type and tosses (!) the
+            # rest of the headers. If that ever changes, this should be a sane
+            # approach.
+            try:
+                data = data.decode('ascii')
+            except UnicodeDecodeError:
+                data = quote(data)
+        try:
+            unquoted = unquote(data, errors='strict')
+        except UnicodeDecodeError:
+            return data
+    return unquoted
+
+
 def quote(value, safe='/'):
     """
     Patched version of urllib.quote that encodes utf8 strings before quoting.
@@ -107,11 +141,7 @@ def quote(value, safe='/'):
     """
     if six.PY3:
         return _quote(value, safe=safe)
-    value = encode_utf8(value)
-    if isinstance(value, bytes):
-        return _quote(value, safe)
-    else:
-        return value
+    return _quote(encode_utf8(value), safe)
 
 
 def encode_utf8(value):
@@ -469,6 +499,14 @@ def get_auth(auth_url, user, key, **kwargs):
         return storage_url, token
 
 
+def resp_header_dict(resp):
+    resp_headers = {}
+    for header, value in resp.getheaders():
+        header = parse_header_string(header).lower()
+        resp_headers[header] = parse_header_string(value)
+    return resp_headers
+
+
 def store_response(resp, response_dict):
     """
     store information about an operation into a dict
@@ -479,17 +517,14 @@ def store_response(resp, response_dict):
        status, reason and a dict of lower-cased headers
     """
     if response_dict is not None:
-        resp_headers = {}
-        for header, value in resp.getheaders():
-            resp_headers[header.lower()] = value
-
         response_dict['status'] = resp.status
         response_dict['reason'] = resp.reason
-        response_dict['headers'] = resp_headers
+        response_dict['headers'] = resp_header_dict(resp)
 
 
 def get_account(url, token, marker=None, limit=None, prefix=None,
-                end_marker=None, http_conn=None, full_listing=False):
+                end_marker=None, http_conn=None, full_listing=False,
+                service_token=None):
     """
     Get a listing of containers for the account.
 
@@ -503,6 +538,7 @@ def get_account(url, token, marker=None, limit=None, prefix=None,
                       conn object)
     :param full_listing: if True, return a full listing, else returns a max
                          of 10000 listings
+    :param service_token: service auth token
     :returns: a tuple of (response headers, a list of containers) The response
               headers will be a dict and all header names will be lowercase.
     :raises ClientException: HTTP GET request failed
@@ -532,15 +568,15 @@ def get_account(url, token, marker=None, limit=None, prefix=None,
         qs += '&end_marker=%s' % quote(end_marker)
     full_path = '%s?%s' % (parsed.path, qs)
     headers = {'X-Auth-Token': token}
+    if service_token:
+        headers['X-Service-Token'] = service_token
     method = 'GET'
     conn.request(method, full_path, '', headers)
     resp = conn.getresponse()
     body = resp.read()
     http_log(("%s?%s" % (url, qs), method,), {'headers': headers}, resp, body)
 
-    resp_headers = {}
-    for header, value in resp.getheaders():
-        resp_headers[header.lower()] = value
+    resp_headers = resp_header_dict(resp)
     if resp.status < 200 or resp.status >= 300:
         raise ClientException('Account GET failed', http_scheme=parsed.scheme,
                               http_host=conn.host, http_path=parsed.path,
@@ -552,7 +588,7 @@ def get_account(url, token, marker=None, limit=None, prefix=None,
     return resp_headers, parse_api_response(resp_headers, body)
 
 
-def head_account(url, token, http_conn=None):
+def head_account(url, token, http_conn=None, service_token=None):
     """
     Get account stats.
 
@@ -560,6 +596,7 @@ def head_account(url, token, http_conn=None):
     :param token: auth token
     :param http_conn: HTTP connection object (If None, it will create the
                       conn object)
+    :param service_token: service auth token
     :returns: a dict containing the response's headers (all header names will
               be lowercase)
     :raises ClientException: HTTP HEAD request failed
@@ -570,6 +607,8 @@ def head_account(url, token, http_conn=None):
         parsed, conn = http_connection(url)
     method = "HEAD"
     headers = {'X-Auth-Token': token}
+    if service_token:
+        headers['X-Service-Token'] = service_token
     conn.request(method, parsed.path, '', headers)
     resp = conn.getresponse()
     body = resp.read()
@@ -579,13 +618,12 @@ def head_account(url, token, http_conn=None):
                               http_host=conn.host, http_path=parsed.path,
                               http_status=resp.status, http_reason=resp.reason,
                               http_response_content=body)
-    resp_headers = {}
-    for header, value in resp.getheaders():
-        resp_headers[header.lower()] = value
+    resp_headers = resp_header_dict(resp)
     return resp_headers
 
 
-def post_account(url, token, headers, http_conn=None, response_dict=None):
+def post_account(url, token, headers, http_conn=None, response_dict=None,
+                 service_token=None):
     """
     Update an account's metadata.
 
@@ -596,6 +634,7 @@ def post_account(url, token, headers, http_conn=None, response_dict=None):
                       conn object)
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :raises ClientException: HTTP POST request failed
     """
     if http_conn:
@@ -604,6 +643,8 @@ def post_account(url, token, headers, http_conn=None, response_dict=None):
         parsed, conn = http_connection(url)
     method = 'POST'
     headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     conn.request(method, parsed.path, '', headers)
     resp = conn.getresponse()
     body = resp.read()
@@ -624,7 +665,7 @@ def post_account(url, token, headers, http_conn=None, response_dict=None):
 def get_container(url, token, container, marker=None, limit=None,
                   prefix=None, delimiter=None, end_marker=None,
                   path=None, http_conn=None,
-                  full_listing=False):
+                  full_listing=False, service_token=None, headers=None):
     """
     Get a listing of objects for the container.
 
@@ -641,15 +682,23 @@ def get_container(url, token, container, marker=None, limit=None,
                       conn object)
     :param full_listing: if True, return a full listing, else returns a max
                          of 10000 listings
+    :param service_token: service auth token
+    :param headers: additional headers to include in the request
     :returns: a tuple of (response headers, a list of objects) The response
               headers will be a dict and all header names will be lowercase.
     :raises ClientException: HTTP GET request failed
     """
     if not http_conn:
         http_conn = http_connection(url)
+    if headers:
+        headers = dict(headers)
+    else:
+        headers = {}
+    headers['X-Auth-Token'] = token
     if full_listing:
         rv = get_container(url, token, container, marker, limit, prefix,
-                           delimiter, end_marker, path, http_conn)
+                           delimiter, end_marker, path, http_conn,
+                           service_token, headers=headers)
         listing = rv[1]
         while listing:
             if not delimiter:
@@ -658,7 +707,8 @@ def get_container(url, token, container, marker=None, limit=None,
                 marker = listing[-1].get('name', listing[-1].get('subdir'))
             listing = get_container(url, token, container, marker, limit,
                                     prefix, delimiter, end_marker, path,
-                                    http_conn)[1]
+                                    http_conn, service_token,
+                                    headers=headers)[1]
             if listing:
                 rv[1].extend(listing)
         return rv
@@ -677,7 +727,8 @@ def get_container(url, token, container, marker=None, limit=None,
         qs += '&end_marker=%s' % quote(end_marker)
     if path:
         qs += '&path=%s' % quote(path)
-    headers = {'X-Auth-Token': token}
+    if service_token:
+        headers['X-Service-Token'] = service_token
     method = 'GET'
     conn.request(method, '%s?%s' % (cont_path, qs), '', headers)
     resp = conn.getresponse()
@@ -694,15 +745,14 @@ def get_container(url, token, container, marker=None, limit=None,
                               http_path=cont_path, http_query=qs,
                               http_status=resp.status, http_reason=resp.reason,
                               http_response_content=body)
-    resp_headers = {}
-    for header, value in resp.getheaders():
-        resp_headers[header.lower()] = value
+    resp_headers = resp_header_dict(resp)
     if resp.status == 204:
         return resp_headers, []
     return resp_headers, parse_api_response(resp_headers, body)
 
 
-def head_container(url, token, container, http_conn=None, headers=None):
+def head_container(url, token, container, http_conn=None, headers=None,
+                   service_token=None):
     """
     Get container stats.
 
@@ -711,6 +761,8 @@ def head_container(url, token, container, http_conn=None, headers=None):
     :param container: container name to get stats for
     :param http_conn: HTTP connection object (If None, it will create the
                       conn object)
+    :param headers: additional headers to include in the request
+    :param service_token: service auth token
     :returns: a dict containing the response's headers (all header names will
               be lowercase)
     :raises ClientException: HTTP HEAD request failed
@@ -722,6 +774,8 @@ def head_container(url, token, container, http_conn=None, headers=None):
     path = '%s/%s' % (parsed.path, quote(container))
     method = 'HEAD'
     req_headers = {'X-Auth-Token': token}
+    if service_token:
+        req_headers['X-Service-Token'] = service_token
     if headers:
         req_headers.update(headers)
     conn.request(method, path, '', req_headers)
@@ -736,14 +790,12 @@ def head_container(url, token, container, http_conn=None, headers=None):
                               http_path=path, http_status=resp.status,
                               http_reason=resp.reason,
                               http_response_content=body)
-    resp_headers = {}
-    for header, value in resp.getheaders():
-        resp_headers[header.lower()] = value
+    resp_headers = resp_header_dict(resp)
     return resp_headers
 
 
 def put_container(url, token, container, headers=None, http_conn=None,
-                  response_dict=None):
+                  response_dict=None, service_token=None):
     """
     Create a container
 
@@ -755,6 +807,7 @@ def put_container(url, token, container, headers=None, http_conn=None,
                       conn object)
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :raises ClientException: HTTP PUT request failed
     """
     if http_conn:
@@ -766,6 +819,8 @@ def put_container(url, token, container, headers=None, http_conn=None,
     if not headers:
         headers = {}
     headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     if 'content-length' not in (k.lower() for k in headers):
         headers['Content-Length'] = '0'
     conn.request(method, path, '', headers)
@@ -785,7 +840,7 @@ def put_container(url, token, container, headers=None, http_conn=None,
 
 
 def post_container(url, token, container, headers, http_conn=None,
-                   response_dict=None):
+                   response_dict=None, service_token=None):
     """
     Update a container's metadata.
 
@@ -797,6 +852,7 @@ def post_container(url, token, container, headers, http_conn=None,
                       conn object)
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :raises ClientException: HTTP POST request failed
     """
     if http_conn:
@@ -806,6 +862,8 @@ def post_container(url, token, container, headers, http_conn=None,
     path = '%s/%s' % (parsed.path, quote(container))
     method = 'POST'
     headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     if 'content-length' not in (k.lower() for k in headers):
         headers['Content-Length'] = '0'
     conn.request(method, path, '', headers)
@@ -825,7 +883,7 @@ def post_container(url, token, container, headers, http_conn=None,
 
 
 def delete_container(url, token, container, http_conn=None,
-                     response_dict=None):
+                     response_dict=None, service_token=None):
     """
     Delete a container
 
@@ -836,6 +894,7 @@ def delete_container(url, token, container, http_conn=None,
                       conn object)
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :raises ClientException: HTTP DELETE request failed
     """
     if http_conn:
@@ -844,6 +903,8 @@ def delete_container(url, token, container, http_conn=None,
         parsed, conn = http_connection(url)
     path = '%s/%s' % (parsed.path, quote(container))
     headers = {'X-Auth-Token': token}
+    if service_token:
+        headers['X-Service-Token'] = service_token
     method = 'DELETE'
     conn.request(method, path, '', headers)
     resp = conn.getresponse()
@@ -863,7 +924,7 @@ def delete_container(url, token, container, http_conn=None,
 
 def get_object(url, token, container, name, http_conn=None,
                resp_chunk_size=None, query_string=None,
-               response_dict=None, headers=None):
+               response_dict=None, headers=None, service_token=None):
     """
     Get an object
 
@@ -882,6 +943,7 @@ def get_object(url, token, container, name, http_conn=None,
                      the response - status, reason and headers
     :param headers: an optional dictionary with additional headers to include
                     in the request
+    :param service_token: service auth token
     :returns: a tuple of (response headers, the object's contents) The response
               headers will be a dict and all header names will be lowercase.
     :raises ClientException: HTTP GET request failed
@@ -896,6 +958,8 @@ def get_object(url, token, container, name, http_conn=None,
     method = 'GET'
     headers = headers.copy() if headers else {}
     headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     conn.request(method, path, '', headers)
     resp = conn.getresponse()
 
@@ -923,7 +987,8 @@ def get_object(url, token, container, name, http_conn=None,
     return parsed_response['headers'], object_body
 
 
-def head_object(url, token, container, name, http_conn=None):
+def head_object(url, token, container, name, http_conn=None,
+                service_token=None, headers=None):
     """
     Get object info
 
@@ -933,6 +998,8 @@ def head_object(url, token, container, name, http_conn=None):
     :param name: object name to get info for
     :param http_conn: HTTP connection object (If None, it will create the
                       conn object)
+    :param service_token: service auth token
+    :param headers: additional headers to include in the request
     :returns: a dict containing the response's headers (all header names will
               be lowercase)
     :raises ClientException: HTTP HEAD request failed
@@ -942,8 +1009,14 @@ def head_object(url, token, container, name, http_conn=None):
     else:
         parsed, conn = http_connection(url)
     path = '%s/%s/%s' % (parsed.path, quote(container), quote(name))
+    if headers:
+        headers = dict(headers)
+    else:
+        headers = {}
+    headers['X-Auth-Token'] = token
     method = 'HEAD'
-    headers = {'X-Auth-Token': token}
+    if service_token:
+        headers['X-Service-Token'] = service_token
     conn.request(method, path, '', headers)
     resp = conn.getresponse()
     body = resp.read()
@@ -954,16 +1027,14 @@ def head_object(url, token, container, name, http_conn=None):
                               http_host=conn.host, http_path=path,
                               http_status=resp.status, http_reason=resp.reason,
                               http_response_content=body)
-    resp_headers = {}
-    for header, value in resp.getheaders():
-        resp_headers[header.lower()] = value
+    resp_headers = resp_header_dict(resp)
     return resp_headers
 
 
 def put_object(url, token=None, container=None, name=None, contents=None,
                content_length=None, etag=None, chunk_size=None,
                content_type=None, headers=None, http_conn=None, proxy=None,
-               query_string=None, response_dict=None):
+               query_string=None, response_dict=None, service_token=None):
     """
     Put an object
 
@@ -973,7 +1044,7 @@ def put_object(url, token=None, container=None, name=None, contents=None,
                       container name is expected to be part of the url
     :param name: object name to put; if None, the object name is expected to be
                  part of the url
-    :param contents: a string, a file like object or an iterable
+    :param contents: a string, a file-like object or an iterable
                      to read object data from;
                      if None, a zero-byte put will be done
     :param content_length: value to send as content-length header; also limits
@@ -984,8 +1055,11 @@ def put_object(url, token=None, container=None, name=None, contents=None,
     :param chunk_size: chunk size of data to write; it defaults to 65536;
                        used only if the contents object has a 'read'
                        method, e.g. file-like objects, ignored otherwise
-    :param content_type: value to send as content-type header; if None, an
-                         empty string value will be sent
+
+    :param content_type: value to send as content-type header, overriding any
+                       value included in the headers param; if None and no
+                       value is found in the headers param, an empty string
+                       value will be sent
     :param headers: additional headers to include in the request, if any
     :param http_conn: HTTP connection object (If None, it will create the
                       conn object)
@@ -994,6 +1068,7 @@ def put_object(url, token=None, container=None, name=None, contents=None,
     :param query_string: if set will be appended with '?' to generated path
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :returns: etag
     :raises ClientException: HTTP PUT request failed
     """
@@ -1014,6 +1089,8 @@ def put_object(url, token=None, container=None, name=None, contents=None,
         headers = {}
     if token:
         headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     if etag:
         headers['ETag'] = etag.strip('"')
     if content_length is not None:
@@ -1024,7 +1101,8 @@ def put_object(url, token=None, container=None, name=None, contents=None,
                 content_length = int(v)
     if content_type is not None:
         headers['Content-Type'] = content_type
-    else:  # python-requests sets application/x-www-form-urlencoded otherwise
+    elif 'Content-Type' not in headers:
+        # python-requests sets application/x-www-form-urlencoded otherwise
         headers['Content-Type'] = ''
     if not contents:
         headers['Content-Length'] = '0'
@@ -1067,7 +1145,7 @@ def put_object(url, token=None, container=None, name=None, contents=None,
 
 
 def post_object(url, token, container, name, headers, http_conn=None,
-                response_dict=None):
+                response_dict=None, service_token=None):
     """
     Update object metadata
 
@@ -1080,6 +1158,7 @@ def post_object(url, token, container, name, headers, http_conn=None,
                       conn object)
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :raises ClientException: HTTP POST request failed
     """
     if http_conn:
@@ -1088,6 +1167,8 @@ def post_object(url, token, container, name, headers, http_conn=None,
         parsed, conn = http_connection(url)
     path = '%s/%s/%s' % (parsed.path, quote(container), quote(name))
     headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     conn.request('POST', path, '', headers)
     resp = conn.getresponse()
     body = resp.read()
@@ -1105,7 +1186,7 @@ def post_object(url, token, container, name, headers, http_conn=None,
 
 def delete_object(url, token=None, container=None, name=None, http_conn=None,
                   headers=None, proxy=None, query_string=None,
-                  response_dict=None):
+                  response_dict=None, service_token=None):
     """
     Delete object
 
@@ -1123,6 +1204,7 @@ def delete_object(url, token=None, container=None, name=None, http_conn=None,
     :param query_string: if set will be appended with '?' to generated path
     :param response_dict: an optional dictionary into which to place
                      the response - status, reason and headers
+    :param service_token: service auth token
     :raises ClientException: HTTP DELETE request failed
     """
     if http_conn:
@@ -1142,6 +1224,8 @@ def delete_object(url, token=None, container=None, name=None, http_conn=None,
         headers = {}
     if token:
         headers['X-Auth-Token'] = token
+    if service_token:
+        headers['X-Service-Token'] = service_token
     conn.request('DELETE', path, '', headers)
     resp = conn.getresponse()
     body = resp.read()
@@ -1177,14 +1261,24 @@ def get_capabilities(http_conn):
                               http_host=conn.host, http_path=parsed.path,
                               http_status=resp.status, http_reason=resp.reason,
                               http_response_content=body)
-    resp_headers = {}
-    for header, value in resp.getheaders():
-        resp_headers[header.lower()] = value
+    resp_headers = resp_header_dict(resp)
     return parse_api_response(resp_headers, body)
 
 
 class Connection(object):
-    """Convenience class to make requests that will also retry the request"""
+
+    """
+    Convenience class to make requests that will also retry the request
+
+    Requests will have an X-Auth-Token header whose value is either
+    the preauthtoken or a token obtained from the auth service using
+    the user credentials provided as args to the constructor. If
+    os_options includes a service_username then requests will also have
+    an X-Service-Token header whose value is a token obtained from the
+    auth service using the service credentials. In this case the request
+    url will be set to the storage_url obtained from the auth service
+    for the service user, unless this is overridden by a preauthurl.
+    """
 
     def __init__(self, authurl=None, user=None, key=None, retries=5,
                  preauthurl=None, preauthtoken=None, snet=False,
@@ -1209,7 +1303,8 @@ class Connection(object):
                             to an auth 2.0 system.
         :param os_options: The OpenStack options which can have tenant_id,
                            auth_token, service_type, endpoint_type,
-                           tenant_name, object_storage_url, region_name
+                           tenant_name, object_storage_url, region_name,
+                           service_username, service_project_name, service_key
         :param insecure: Allow to access servers without checking SSL certs.
                          The server's certificate will not be verified.
         :param ssl_compression: Whether to enable compression at the SSL layer.
@@ -1233,13 +1328,18 @@ class Connection(object):
         self.starting_backoff = starting_backoff
         self.max_backoff = max_backoff
         self.auth_version = auth_version
-        self.os_options = os_options or {}
+        self.os_options = dict(os_options or {})
         if tenant_name:
             self.os_options['tenant_name'] = tenant_name
         if preauthurl:
             self.os_options['object_storage_url'] = preauthurl
         self.url = preauthurl or self.os_options.get('object_storage_url')
         self.token = preauthtoken or self.os_options.get('auth_token')
+        if self.os_options.get('service_username', None):
+            self.service_auth = True
+        else:
+            self.service_auth = False
+        self.service_token = None
         self.cacert = cacert
         self.insecure = insecure
         self.ssl_compression = ssl_compression
@@ -1258,10 +1358,29 @@ class Connection(object):
                 self.http_conn = None
 
     def get_auth(self):
-        return get_auth(self.authurl, self.user, self.key,
+        self.url, self.token = get_auth(self.authurl, self.user, self.key,
+                                        snet=self.snet,
+                                        auth_version=self.auth_version,
+                                        os_options=self.os_options,
+                                        cacert=self.cacert,
+                                        insecure=self.insecure,
+                                        timeout=self.timeout)
+        return self.url, self.token
+
+    def get_service_auth(self):
+        opts = self.os_options
+        service_options = {}
+        service_options['tenant_name'] = opts.get('service_project_name', None)
+        service_options['region_name'] = opts.get('region_name', None)
+        service_options['object_storage_url'] = opts.get('object_storage_url',
+                                                         None)
+        service_user = opts.get('service_username', None)
+        service_key = opts.get('service_key', None)
+        return get_auth(self.authurl, service_user,
+                        service_key,
                         snet=self.snet,
                         auth_version=self.auth_version,
-                        os_options=self.os_options,
+                        os_options=service_options,
                         cacert=self.cacert,
                         insecure=self.insecure,
                         timeout=self.timeout)
@@ -1293,13 +1412,17 @@ class Connection(object):
                 if not self.url or not self.token:
                     self.url, self.token = self.get_auth()
                     self.http_conn = None
+                if self.service_auth and not self.service_token:
+                    self.url, self.service_token = self.get_service_auth()
+                    self.http_conn = None
                 self.auth_end_time = time()
                 if not self.http_conn:
                     self.http_conn = self.http_connection()
                 kwargs['http_conn'] = self.http_conn
                 if caller_response_dict is not None:
                     kwargs['response_dict'] = {}
-                rv = func(self.url, self.token, *args, **kwargs)
+                rv = func(self.url, self.token, *args,
+                          service_token=self.service_token, **kwargs)
                 self._add_response_dict(caller_response_dict, kwargs)
                 return rv
             except SSLError:
@@ -1312,11 +1435,11 @@ class Connection(object):
                 self.http_conn = None
             except ClientException as err:
                 self._add_response_dict(caller_response_dict, kwargs)
-                if self.attempts > self.retries:
+                if self.attempts > self.retries or err.http_status is None:
                     logger.exception(err)
                     raise
                 if err.http_status == 401:
-                    self.url = self.token = None
+                    self.url = self.token = self.service_token = None
                     if retried_auth or not all((self.authurl,
                                                 self.user,
                                                 self.key)):
@@ -1356,13 +1479,13 @@ class Connection(object):
         return self._retry(None, post_account, headers,
                            response_dict=response_dict)
 
-    def head_container(self, container):
+    def head_container(self, container, headers=None):
         """Wrapper for :func:`head_container`"""
-        return self._retry(None, head_container, container)
+        return self._retry(None, head_container, container, headers=headers)
 
     def get_container(self, container, marker=None, limit=None, prefix=None,
                       delimiter=None, end_marker=None, path=None,
-                      full_listing=False):
+                      full_listing=False, headers=None):
         """Wrapper for :func:`get_container`"""
         # TODO(unknown): With full_listing=True this will restart the entire
         # listing with each retry. Need to make a better version that just
@@ -1370,7 +1493,7 @@ class Connection(object):
         return self._retry(None, get_container, container, marker=marker,
                            limit=limit, prefix=prefix, delimiter=delimiter,
                            end_marker=end_marker, path=path,
-                           full_listing=full_listing)
+                           full_listing=full_listing, headers=headers)
 
     def put_container(self, container, headers=None, response_dict=None):
         """Wrapper for :func:`put_container`"""
@@ -1387,9 +1510,9 @@ class Connection(object):
         return self._retry(None, delete_container, container,
                            response_dict=response_dict)
 
-    def head_object(self, container, obj):
+    def head_object(self, container, obj, headers=None):
         """Wrapper for :func:`head_object`"""
-        return self._retry(None, head_object, container, obj)
+        return self._retry(None, head_object, container, obj, headers=headers)
 
     def get_object(self, container, obj, resp_chunk_size=None,
                    query_string=None, response_dict=None, headers=None):
