@@ -12,7 +12,9 @@
 # implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+import logging
 import os
+
 from concurrent.futures import as_completed, CancelledError, TimeoutError
 from copy import deepcopy
 from errno import EEXIST, ENOENT
@@ -27,7 +29,7 @@ from threading import Thread
 from six import StringIO, text_type
 from six.moves.queue import Queue
 from six.moves.queue import Empty as QueueEmpty
-from six.moves.urllib.parse import quote, unquote
+from six.moves.urllib.parse import quote
 from six import Iterator, string_types
 
 import json
@@ -39,12 +41,15 @@ from swiftclient.command_helpers import (
 )
 from swiftclient.utils import (
     config_true_value, ReadableToIterable, LengthWrapper, EMPTY_ETAG,
-    parse_api_response
+    parse_api_response, report_traceback
 )
 from swiftclient.exceptions import ClientException
 from swiftclient.multithreading import MultiThreadingManager
 
 
+logger = logging.getLogger("swiftclient.service")
+
+
 class ResultsIterator(Iterator):
     def __init__(self, futures):
         self.futures = interruptable_as_completed(futures)
@@ -176,7 +181,8 @@ _default_local_options = {
     'fail_fast': False,
     'human': False,
     'dir_marker': False,
-    'checksum': True
+    'checksum': True,
+    'shuffle': False
 }
 
 POLICY = 'X-Storage-Policy'
@@ -434,16 +440,24 @@ class SwiftService(object):
                     return res
                 except ClientException as err:
                     if err.http_status != 404:
+                        traceback, err_time = report_traceback()
+                        logger.exception(err)
                         res.update({
                             'success': False,
-                            'error': err
+                            'error': err,
+                            'traceback': traceback,
+                            'error_timestamp': err_time
                         })
                         return res
                     raise SwiftError('Account not found', exc=err)
                 except Exception as err:
+                    traceback, err_time = report_traceback()
+                    logger.exception(err)
                     res.update({
                         'success': False,
-                        'error': err
+                        'error': err,
+                        'traceback': traceback,
+                        'error_timestamp': err_time
                     })
                     return res
         else:
@@ -466,17 +480,25 @@ class SwiftService(object):
                     return res
                 except ClientException as err:
                     if err.http_status != 404:
+                        traceback, err_time = report_traceback()
+                        logger.exception(err)
                         res.update({
                             'success': False,
-                            'error': err
+                            'error': err,
+                            'traceback': traceback,
+                            'error_timestamp': err_time
                         })
                         return res
                     raise SwiftError('Container %r not found' % container,
                                      container=container, exc=err)
                 except Exception as err:
+                    traceback, err_time = report_traceback()
+                    logger.exception(err)
                     res.update({
                         'success': False,
-                        'error': err
+                        'error': err,
+                        'traceback': traceback,
+                        'error_timestamp': err_time
                     })
                     return res
             else:
@@ -505,9 +527,13 @@ class SwiftService(object):
             })
             return res
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res.update({
                 'success': False,
-                'error': err
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
             })
             return res
 
@@ -580,18 +606,26 @@ class SwiftService(object):
                     get_future_result(post)
                 except ClientException as err:
                     if err.http_status != 404:
+                        traceback, err_time = report_traceback()
+                        logger.exception(err)
                         res.update({
                             'success': False,
                             'error': err,
+                            'traceback': traceback,
+                            'error_timestamp': err_time,
                             'response_dict': response_dict
                         })
                         return res
-                    raise SwiftError('Account not found')
+                    raise SwiftError('Account not found', exc=err)
                 except Exception as err:
+                    traceback, err_time = report_traceback()
+                    logger.exception(err)
                     res.update({
                         'success': False,
                         'error': err,
-                        'response_dict': response_dict
+                        'response_dict': response_dict,
+                        'traceback': traceback,
+                        'error_timestamp': err_time
                     })
             return res
         if not objects:
@@ -618,23 +652,31 @@ class SwiftService(object):
                 get_future_result(post)
             except ClientException as err:
                 if err.http_status != 404:
+                    traceback, err_time = report_traceback()
+                    logger.exception(err)
                     res.update({
                         'action': 'post_container',
                         'success': False,
                         'error': err,
+                        'traceback': traceback,
+                        'error_timestamp': err_time,
                         'response_dict': response_dict
                     })
                     return res
                 raise SwiftError(
                     "Container '%s' not found" % container,
-                    container=container
+                    container=container, exc=err
                 )
             except Exception as err:
+                traceback, err_time = report_traceback()
+                logger.exception(err)
                 res.update({
                     'action': 'post_container',
                     'success': False,
                     'error': err,
-                    'response_dict': response_dict
+                    'response_dict': response_dict,
+                    'traceback': traceback,
+                    'error_timestamp': err_time
                 })
             return res
         else:
@@ -719,9 +761,13 @@ class SwiftService(object):
             conn.post_object(
                 container, obj, headers=headers, response_dict=result)
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res.update({
                 'success': False,
-                'error': err
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
             })
 
         return res
@@ -752,7 +798,7 @@ class SwiftService(object):
         else:
             options = self._options
 
-        rq = Queue()
+        rq = Queue(maxsize=10)  # Just stop list running away consuming memory
 
         if container is None:
             listing_future = self.thread_manager.container_pool.submit(
@@ -774,7 +820,6 @@ class SwiftService(object):
     @staticmethod
     def _list_account_job(conn, options, result_queue):
         marker = ''
-        success = True
         error = None
         try:
             while True:
@@ -803,23 +848,30 @@ class SwiftService(object):
 
                 marker = items[-1].get('name', items[-1].get('subdir'))
         except ClientException as err:
-            success = False
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             if err.http_status != 404:
-                error = err
+                error = (err, traceback, err_time)
             else:
-                error = SwiftError('Account not found')
+                error = (
+                    SwiftError('Account not found', exc=err),
+                    traceback, err_time
+                )
 
         except Exception as err:
-            success = False
-            error = err
+            traceback, err_time = report_traceback()
+            logger.exception(err)
+            error = (err, traceback, err_time)
 
         res = {
             'action': 'list_account_part',
             'container': None,
             'prefix': options['prefix'],
-            'success': success,
+            'success': False,
             'marker': marker,
-            'error': error,
+            'error': error[0],
+            'traceback': error[1],
+            'error_timestamp': error[2]
         }
         result_queue.put(res)
         result_queue.put(None)
@@ -827,7 +879,6 @@ class SwiftService(object):
     @staticmethod
     def _list_container_job(conn, container, options, result_queue):
         marker = ''
-        success = True
         error = None
         try:
             while True:
@@ -852,23 +903,33 @@ class SwiftService(object):
 
                 marker = items[-1].get('name', items[-1].get('subdir'))
         except ClientException as err:
-            success = False
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             if err.http_status != 404:
-                error = err
+                error = (err, traceback, err_time)
             else:
-                error = SwiftError('Container %r not found' % container,
-                                   container=container)
+                error = (
+                    SwiftError(
+                        'Container %r not found' % container,
+                        container=container, exc=err
+                    ),
+                    traceback,
+                    err_time
+                )
         except Exception as err:
-            success = False
-            error = err
+            traceback, err_time = report_traceback()
+            logger.exception(err)
+            error = (err, traceback, err_time)
 
         res = {
             'action': 'list_container_part',
             'container': container,
             'prefix': options['prefix'],
-            'success': success,
+            'success': False,
             'marker': marker,
-            'error': error,
+            'error': error[0],
+            'traceback': error[1],
+            'error_timestamp': error[2]
         }
         result_queue.put(res)
         result_queue.put(None)
@@ -895,6 +956,7 @@ class SwiftService(object):
                                 'out_directory': None,
                                 'out_file': None,
                                 'remove_prefix': False,
+                                'shuffle' : False
                             }
 
         :returns: A generator for returning the results of the download
@@ -916,45 +978,26 @@ class SwiftService(object):
                 try:
                     options_copy = deepcopy(options)
                     options_copy["long"] = False
-                    containers = []
+
                     for part in self.list(options=options_copy):
                         if part["success"]:
-                            containers.extend([
-                                i['name'] for i in part["listing"]
-                            ])
+                            containers = [i['name'] for i in part["listing"]]
+
+                            if options['shuffle']:
+                                shuffle(containers)
+
+                            for con in containers:
+                                for res in self._download_container(
+                                        con, options_copy):
+                                    yield res
                         else:
                             raise part["error"]
 
-                    shuffle(containers)
-
-                    o_downs = []
-                    for con in containers:
-                        objs = []
-                        for part in self.list(
-                                container=con, options=options_copy):
-                            if part["success"]:
-                                objs.extend([
-                                    i['name'] for i in part["listing"]
-                                ])
-                            else:
-                                raise part["error"]
-                        shuffle(objs)
-
-                        o_downs.extend(
-                            self.thread_manager.object_dd_pool.submit(
-                                self._download_object_job, con, obj,
-                                options_copy
-                            ) for obj in objs
-                        )
-
-                    for o_down in interruptable_as_completed(o_downs):
-                        yield o_down.result()
-
                 # If we see a 404 here, the listing of the account failed
                 except ClientException as err:
                     if err.http_status != 404:
                         raise
-                    raise SwiftError('Account not found')
+                    raise SwiftError('Account not found', exc=err)
 
         elif not objects:
             if '/' in container:
@@ -1140,12 +1183,16 @@ class SwiftService(object):
             return res
 
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res = {
                 'action': 'download_object',
                 'container': container,
                 'object': obj,
                 'success': False,
                 'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time,
                 'response_dict': results_dict,
                 'path': path,
                 'pseudodir': pseudodir,
@@ -1153,14 +1200,17 @@ class SwiftService(object):
             }
             return res
 
-    def _download_container(self, container, options):
+    def _submit_page_downloads(self, container, page_generator, options):
         try:
-            objects = []
-            for part in self.list(container=container, options=options):
-                if part["success"]:
-                    objects.extend([o["name"] for o in part["listing"]])
-                else:
-                    raise part["error"]
+            list_page = next(page_generator)
+        except StopIteration:
+            return None
+
+        if list_page["success"]:
+            objects = [o["name"] for o in list_page["listing"]]
+
+            if options["shuffle"]:
+                shuffle(objects)
 
             o_downs = [
                 self.thread_manager.object_dd_pool.submit(
@@ -1168,14 +1218,62 @@ class SwiftService(object):
                 ) for obj in objects
             ]
 
-            for o_down in interruptable_as_completed(o_downs):
-                yield o_down.result()
+            return o_downs
+        else:
+            raise list_page["error"]
 
+    def _download_container(self, container, options):
+        _page_generator = self.list(container=container, options=options)
+        try:
+            next_page_downs = self._submit_page_downloads(
+                container, _page_generator, options
+            )
         except ClientException as err:
             if err.http_status != 404:
                 raise
-            raise SwiftError('Container %r not found' % container,
-                             container=container)
+            raise SwiftError(
+                'Container %r not found' % container,
+                container=container, exc=err
+            )
+
+        error = None
+        while next_page_downs:
+            page_downs = next_page_downs
+            next_page_downs = None
+
+            # Start downloading the next page of list results when
+            # we have completed 80% of the previous page
+            next_page_triggered = False
+            next_page_trigger_point = 0.8 * len(page_downs)
+
+            page_results_yielded = 0
+            for o_down in interruptable_as_completed(page_downs):
+                yield o_down.result()
+
+                # Do we need to start the next set of downloads yet?
+                if not next_page_triggered:
+                    page_results_yielded += 1
+                    if page_results_yielded >= next_page_trigger_point:
+                        try:
+                            next_page_downs = self._submit_page_downloads(
+                                container, _page_generator, options
+                            )
+                        except ClientException as err:
+                            # Allow the current page to finish downloading
+                            logger.exception(err)
+                            error = err
+                        except Exception:
+                            # Something unexpected went wrong - cancel
+                            # remaining downloads
+                            for _d in page_downs:
+                                _d.cancel()
+                            raise
+                        finally:
+                            # Stop counting and testing
+                            next_page_triggered = True
+
+        if error:
+            raise error
 
     # Upload related methods
     #
@@ -1200,10 +1298,10 @@ class SwiftService(object):
 
                         The SwiftUploadObject source may be one of:
 
-                            file - A file like object (with a read method)
-                            path - A string containing the path to a local file
-                                   or directory
-                            None - Indicates that we want an empty object
+                            * A file-like object (with a read method)
+                            * A string containing the path to a local
+                              file or directory
+                            * None, to indicate that we want an empty object
 
         :param options: A dictionary containing options to override the global
                         options specified during the service object creation.
@@ -1336,12 +1434,16 @@ class SwiftService(object):
                         file_jobs[file_future] = details
                     except OSError as err:
                         # Avoid tying up threads with jobs that will fail
+                        traceback, err_time = report_traceback()
+                        logger.exception(err)
                         res = {
                             'action': 'upload_object',
                             'container': container,
                             'object': o,
                             'success': False,
                             'error': err,
+                            'traceback': traceback,
+                            'error_timestamp': err_time,
                             'path': s
                         }
                         rq.put(res)
@@ -1440,9 +1542,13 @@ class SwiftService(object):
                 'response_dict': create_response
             })
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res.update({
                 'success': False,
                 'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time,
                 'response_dict': create_response
             })
         return res
@@ -1481,9 +1587,14 @@ class SwiftService(object):
                     return res
             except ClientException as err:
                 if err.http_status != 404:
+                    traceback, err_time = report_traceback()
+                    logger.exception(err)
                     res.update({
                         'success': False,
-                        'error': err})
+                        'error': err,
+                        'traceback': traceback,
+                        'error_timestamp': err_time
+                    })
                     return res
         try:
             conn.put_object(container, obj, '', content_length=0,
@@ -1495,9 +1606,13 @@ class SwiftService(object):
                 'response_dict': results_dict})
             return res
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res.update({
                 'success': False,
                 'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time,
                 'response_dict': results_dict})
             return res
 
@@ -1550,9 +1665,13 @@ class SwiftService(object):
             return res
 
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res.update({
                 'success': False,
                 'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time,
                 'response_dict': results_dict,
                 'attempts': conn.attempts
             })
@@ -1681,9 +1800,13 @@ class SwiftService(object):
                                 old_slo_manifest_paths.append(seg_path)
                 except ClientException as err:
                     if err.http_status != 404:
+                        traceback, err_time = report_traceback()
+                        logger.exception(err)
                         res.update({
                             'success': False,
-                            'error': err
+                            'error': err,
+                            'traceback': traceback,
+                            'error_timestamp': err_time
                         })
                         return res
 
@@ -1738,9 +1861,11 @@ class SwiftService(object):
                         if not r['success']:
                             errors = True
                         segment_results.append(r)
-                    except Exception as e:
+                    except Exception as err:
+                        traceback, err_time = report_traceback()
+                        logger.exception(err)
                         errors = True
-                        exceptions.append(e)
+                        exceptions.append((err, traceback, err_time))
                 if errors:
                     err = ClientException(
                         'Aborting manifest creation '
@@ -1829,8 +1954,7 @@ class SwiftService(object):
                 delobjsmap = {}
                 if old_manifest:
                     scontainer, sprefix = old_manifest.split('/', 1)
-                    scontainer = unquote(scontainer)
-                    sprefix = unquote(sprefix).rstrip('/') + '/'
+                    sprefix = sprefix.rstrip('/') + '/'
                     delobjsmap[scontainer] = []
                     for part in self.list(scontainer, {'prefix': sprefix}):
                         if not part["success"]:
@@ -1869,16 +1993,26 @@ class SwiftService(object):
             return res
 
         except OSError as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             if err.errno == ENOENT:
-                err = SwiftError('Local file %r not found' % path)
+                error = SwiftError('Local file %r not found' % path, exc=err)
+            else:
+                error = err
             res.update({
                 'success': False,
-                'error': err
+                'error': error,
+                'traceback': traceback,
+                'error_timestamp': err_time
             })
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             res.update({
                 'success': False,
-                'error': err
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
             })
         return res
 
@@ -1977,8 +2111,15 @@ class SwiftService(object):
         try:
             conn.delete_object(container, obj, response_dict=results_dict)
             res = {'success': True}
-        except Exception as e:
-            res = {'success': False, 'error': e}
+        except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
+            res = {
+                'success': False,
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
+            }
 
         res.update({
             'action': 'delete_segment',
@@ -1994,12 +2135,12 @@ class SwiftService(object):
 
     def _delete_object(self, conn, container, obj, options,
                        results_queue=None):
+        res = {
+            'action': 'delete_object',
+            'container': container,
+            'object': obj
+        }
         try:
-            res = {
-                'action': 'delete_object',
-                'container': container,
-                'object': obj
-            }
             old_manifest = None
             query_string = None
 
@@ -2022,8 +2163,7 @@ class SwiftService(object):
                 dlo_segments_deleted = True
                 segment_pool = self.thread_manager.segment_pool
                 s_container, s_prefix = old_manifest.split('/', 1)
-                s_container = unquote(s_container)
-                s_prefix = unquote(s_prefix).rstrip('/') + '/'
+                s_prefix = s_prefix.rstrip('/') + '/'
 
                 del_segs = []
                 for part in self.list(
@@ -2054,8 +2194,14 @@ class SwiftService(object):
             })
 
         except Exception as err:
-            res['success'] = False
-            res['error'] = err
+            traceback, err_time = report_traceback()
+            logger.exception(err)
+            res.update({
+                'success': False,
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
+            })
             return res
 
         return res
@@ -2066,8 +2212,15 @@ class SwiftService(object):
         try:
             conn.delete_container(container, response_dict=results_dict)
             res = {'success': True}
-        except Exception as e:
-            res = {'success': False, 'error': e}
+        except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
+            res = {
+                'success': False,
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
+            }
 
         res.update({
             'action': 'delete_container',
@@ -2080,29 +2233,34 @@ class SwiftService(object):
 
     def _delete_container(self, container, options):
         try:
-            objs = []
             for part in self.list(container=container):
                 if part["success"]:
-                    objs.extend([o['name'] for o in part['listing']])
+                    objs = [o['name'] for o in part['listing']]
+
+                    o_dels = self.delete(
+                        container=container, objects=objs, options=options
+                    )
+                    for res in o_dels:
+                        yield res
                 else:
                     raise part["error"]
 
-            for res in self.delete(
-                    container=container, objects=objs, options=options):
-                yield res
-
             con_del = self.thread_manager.container_pool.submit(
                 self._delete_empty_container, container
             )
             con_del_res = get_future_result(con_del)
 
         except Exception as err:
+            traceback, err_time = report_traceback()
+            logger.exception(err)
             con_del_res = {
                 'action': 'delete_container',
                 'container': container,
                 'object': None,
                 'success': False,
-                'error': err
+                'error': err,
+                'traceback': traceback,
+                'error_timestamp': err_time
             }
 
         yield con_del_res
@@ -2140,7 +2298,7 @@ class SwiftService(object):
         except ClientException as err:
             if err.http_status != 404:
                 raise err
-            raise SwiftError('Account not found')
+            raise SwiftError('Account not found', exc=err)
 
         return res
 
@@ -2175,10 +2333,16 @@ class SwiftService(object):
                 res['status'] = 'cancelled'
                 result_queue.put(res)
             except Exception as err:
+                traceback, err_time = report_traceback()
+                logger.exception(err)
                 details = futures[f]
                 res = details
-                res['success'] = False
-                res['error'] = err
+                res.update({
+                    'success': False,
+                    'error': err,
+                    'traceback': traceback,
+                    'error_timestamp': err_time
+                })
                 result_queue.put(res)
 
         result_queue.put(None)
