@@ -51,22 +51,22 @@ def immediate_exit(signum, frame):
     stderr.write(" Aborted\n")
     os_exit(2)
 
-st_delete_options = '''[-all] [--leave-segments]
+st_delete_options = '''[--all] [--leave-segments]
                     [--object-threads <threads>]
                     [--container-threads <threads>]
-                    <container> [object]
+                    [<container>] [<object>] [...]
 '''
 
 st_delete_help = '''
 Delete a container or objects within a container.
 
 Positional arguments:
-  <container>           Name of container to delete from.
-  [object]              Name of object to delete. Specify multiple times
+  [<container>]         Name of container to delete from.
+  [<object>]            Name of object to delete. Specify multiple times
                         for multiple objects.
 
 Optional arguments:
-  --all                 Delete all containers and objects.
+  -a, --all             Delete all containers and objects.
   --leave-segments      Do not delete segments of manifest objects.
   --object-threads <threads>
                         Number of threads to use for deleting objects.
@@ -89,10 +89,10 @@ def st_delete(parser, args, output_manager):
         '', '--object-threads', type=int,
         default=10, help='Number of threads to use for deleting objects. '
         'Default is 10.')
-    parser.add_option('', '--container-threads', type=int,
-                      default=10, help='Number of threads to use for '
-                      'deleting containers. '
-                      'Default is 10.')
+    parser.add_option(
+        '', '--container-threads', type=int,
+        default=10, help='Number of threads to use for deleting containers. '
+        'Default is 10.')
     (options, args) = parse_args(parser, args)
     args = args[1:]
     if (not args and not options.yes_all) or (args and options.yes_all):
@@ -112,8 +112,8 @@ def st_delete(parser, args, output_manager):
                 if '/' in container:
                     output_manager.error(
                         'WARNING: / in container name; you '
-                        "might have meant '%s' instead of '%s'." % (
-                        container.replace('/', ' ', 1), container)
+                        "might have meant '%s' instead of '%s'." %
+                        (container.replace('/', ' ', 1), container)
                     )
                     return
                 objects = args[1:]
@@ -156,6 +156,7 @@ st_download_options = '''[--all] [--marker] [--prefix <prefix>]
                       [--object-threads <threads>]
                       [--container-threads <threads>] [--no-download]
                       [--skip-identical] [--remove-prefix]
+                      [--header <header:value>] [--no-shuffle]
                       <container> <object>
 '''
 
@@ -170,17 +171,18 @@ Positional arguments:
                         objects from the container.
 
 Optional arguments:
-  --all                 Indicates that you really want to download
+  -a, --all             Indicates that you really want to download
                         everything in the account.
-  --marker              Marker to use when starting a container or account
+  -m, --marker          Marker to use when starting a container or account
                         download.
-  --prefix <prefix>     Only download items beginning with <prefix>
-  --remove-prefix       An optional flag for --prefix <prefix>, use this
+  -p, --prefix <prefix> Only download items beginning with <prefix>
+  -r, --remove-prefix   An optional flag for --prefix <prefix>, use this
                         option to download items without <prefix>
-  --output <out_file>   For a single file download, stream the output to
+  -o, --output <out_file>
+                        For a single file download, stream the output to
                         <out_file>. Specifying "-" as <out_file> will
                         redirect to stdout.
-  --output-dir <out_directory>
+  -D, --output-dir <out_directory>
                         An optional directory to which to store objects.
                         By default, all objects are recreated in the current
                         directory.
@@ -192,12 +194,20 @@ Optional arguments:
                         Default is 10.
   --no-download         Perform download(s), but don't actually write anything
                         to disk.
-  --header <header_name:header_value>
+  -H, --header <header:value>
                         Adds a customized request header to the query, like
-                        "Range" or "If-Match". This argument is repeatable.
+                        "Range" or "If-Match". This option may be repeated.
                         Example --header "content-type:text/plain"
   --skip-identical      Skip downloading files that are identical on both
                         sides.
+  --no-shuffle          By default, when downloading a complete account or
+                        container, download order is randomised in order to
+                        to reduce the load on individual drives when multiple
+                        clients are executed simultaneously to download the
+                        same set of objects (e.g. a nightly automated download
+                        script to multiple servers). Enable this option to
+                        submit download jobs to the thread pool in the order
+                        they are listed in the object store.
 '''.strip("\n")
 
 
@@ -241,12 +251,20 @@ def st_download(parser, args, output_manager):
         '-H', '--header', action='append', dest='header',
         default=[],
         help='Adds a customized request header to the query, like "Range" or '
-        '"If-Match". This argument is repeatable. '
+        '"If-Match". This option may be repeated. '
         'Example: --header "content-type:text/plain"')
     parser.add_option(
         '--skip-identical', action='store_true', dest='skip_identical',
         default=False, help='Skip downloading files that are identical on '
         'both sides.')
+    parser.add_option(
+        '--no-shuffle', action='store_false', dest='shuffle',
+        default=True, help='By default, download order is randomised in order '
+        'to reduce the load on individual drives when multiple clients are '
+        'executed simultaneously to download the same set of objects (e.g. a '
+        'nightly automated download script to multiple servers). Enable this '
+        'option to submit download jobs to the thread pool in the order they '
+        'are listed in the object store.')
     (options, args) = parse_args(parser, args)
     args = args[1:]
     if options.out_file == '-':
@@ -277,8 +295,8 @@ def st_download(parser, args, output_manager):
                 if '/' in container:
                     output_manager.error(
                         'WARNING: / in container name; you '
-                        "might have meant '%s' instead of '%s'." % (
-                        container.replace('/', ' ', 1), container)
+                        "might have meant '%s' instead of '%s'." %
+                        (container.replace('/', ' ', 1), container)
                     )
                     return
                 objects = args[1:]
@@ -353,10 +371,12 @@ def st_download(parser, args, output_manager):
 
         except SwiftError as e:
             output_manager.error(e.value)
+        except Exception as e:
+            output_manager.error(e)
 
 
 st_list_options = '''[--long] [--lh] [--totals] [--prefix <prefix>]
-                  [--delimiter <delimiter>]
+                  [--delimiter <delimiter>] [container]
 '''
 
 st_list_help = '''
@@ -366,12 +386,14 @@ Positional arguments:
   [container]           Name of container to list object in.
 
 Optional arguments:
-  --long                Long listing format, similar to ls -l.
+  -l, --long            Long listing format, similar to ls -l.
   --lh                  Report sizes in human readable format similar to
                         ls -lh.
-  --totals              Used with -l or --lh, only report totals.
-  --prefix              Only list items beginning with the prefix.
-  --delimiter           Roll up items with the given delimiter. For containers
+  -t, --totals          Used with -l or --lh, only report totals.
+  -p <prefix>, --prefix <prefix>
+                        Only list items beginning with the prefix.
+  -d <delim>, --delimiter <delim>
+                        Roll up items with the given delimiter. For containers
                         only. See OpenStack Swift API documentation for what
                         this means.
 '''.strip('\n')
@@ -404,6 +426,7 @@ def st_list(parser, args, output_manager):
                             datestamp, item_name)
                 else:    # list container contents
                     subdir = item.get('subdir')
+                    content_type = item.get('content_type')
                     if subdir is None:
                         item_bytes = item.get('bytes')
                         byte_str = prt_bytes(item_bytes, options.human)
@@ -416,7 +439,8 @@ def st_list(parser, args, output_manager):
                         item_name = subdir
                     if not options.totals:
                         output_manager.print_msg(
-                            "%s %10s %8s %s", byte_str, date, xtime, item_name)
+                            "%s %10s %8s %24s %s",
+                            byte_str, date, xtime, content_type, item_name)
                 total_bytes += item_bytes
 
         # report totals
@@ -577,17 +601,22 @@ Positional arguments:
   [object]              Name of object to post.
 
 Optional arguments:
-  --read-acl <acl>      Read ACL for containers. Quick summary of ACL syntax:
+  -r, --read-acl <acl>  Read ACL for containers. Quick summary of ACL syntax:
                         .r:*, .r:-.example.com, .r:www.example.com, account1,
                         account2:user2
-  --write-acl <acl>     Write ACL for containers. Quick summary of ACL syntax:
+  -w, --write-acl <acl> Write ACL for containers. Quick summary of ACL syntax:
                         account1 account2:user2
-  --sync-to <sync-to>   Sync To for containers, for multi-cluster replication.
-  --sync-key <sync-key> Sync Key for containers, for multi-cluster replication.
-  --meta <name:value>   Sets a meta data item. This option may be repeated.
+  -t, --sync-to <sync-to>
+                        Sync To for containers, for multi-cluster replication.
+  -k, --sync-key <sync-key>
+                        Sync Key for containers, for multi-cluster replication.
+  -m, --meta <name:value>
+                        Sets a meta data item. This option may be repeated.
                         Example: -m Color:Blue -m Size:Large
-  --header <header>     Set request headers. This option may be repeated.
-                        Example -H "content-type:text/plain"
+  -H, --header <header:value>
+                        Adds a customized request header.
+                        This option may be repeated. Example
+                        -H "content-type:text/plain" -H "Content-Length: 4000"
 '''.strip('\n')
 
 
@@ -612,7 +641,8 @@ def st_post(parser, args, output_manager):
         'Example: -m Color:Blue -m Size:Large')
     parser.add_option(
         '-H', '--header', action='append', dest='header',
-        default=[], help='Set request headers. This option may be repeated. '
+        default=[], help='Adds a customized request header. '
+        'This option may be repeated. '
         'Example: -H "content-type:text/plain" '
         '-H "Content-Length: 4000"')
     (options, args) = parse_args(parser, args)
@@ -662,11 +692,10 @@ st_upload_options = '''[--changed] [--skip-identical] [--segment-size <size>]
                     [--object-threads <thread>] [--segment-threads <threads>]
                     [--header <header>] [--use-slo] [--ignore-checksum]
                     [--object-name <object-name>]
-                    <container> <file_or_directory>
+                    <container> <file_or_directory> [<file_or_directory>] [...]
 '''
 
-st_upload_help = '''
-Uploads specified files and directories to the given container.
+st_upload_help = ''' Uploads specified files and directories to the given container.
 
 Positional arguments:
   <container>           Name of container to upload to.
@@ -674,10 +703,11 @@ Positional arguments:
                         times for multiple uploads.
 
 Optional arguments:
-  --changed             Only upload files that have changed since the last
+  -c, --changed         Only upload files that have changed since the last
                         upload.
   --skip-identical      Skip uploading files that are identical on both sides.
-  --segment-size <size> Upload files in segments no larger than <size> (in
+  -S, --segment-size <size>
+                        Upload files in segments no larger than <size> (in
                         Bytes) and then create a "manifest" file that will
                         download all the segments as if it were the original
                         file.
@@ -694,9 +724,10 @@ Optional arguments:
   --segment-threads <threads>
                         Number of threads to use for uploading object segments.
                         Default is 10.
-  --header <header>     Set request headers with the syntax header:value.
-                        This option may be repeated.
-                        Example -H "content-type:text/plain".
+  -H, --header <header:value>
+                        Adds a customized request header. This option may be
+                        repeated. Example -H "content-type:text/plain"
+                         -H "Content-Length: 4000".
   --use-slo             When used in conjunction with --segment-size it will
                         create a Static Large Object instead of the default
                         Dynamic Large Object.
@@ -977,7 +1008,8 @@ def st_auth(parser, args, thread_manager):
         print('export OS_AUTH_TOKEN=%s' % sh_quote(token))
 
 
-st_tempurl_options = '<method> <seconds> <path> <key>'
+st_tempurl_options = '''[--absolute]
+                     <method> <seconds> <path> <key>'''
 
 
 st_tempurl_help = '''
@@ -986,17 +1018,30 @@ Generates a temporary URL for a Swift object.
 Positional arguments:
   <method>              An HTTP method to allow for this temporary URL.
                         Usually 'GET' or 'PUT'.
-  <seconds>             The amount of time in seconds the temporary URL will
-                        be valid for.
+  <seconds>             The amount of time in seconds the temporary URL will be
+                        valid for; or, if --absolute is passed, the Unix
+                        timestamp when the temporary URL will expire.
   <path>                The full path to the Swift object. Example:
                         /v1/AUTH_account/c/o.
   <key>                 The secret temporary URL key set on the Swift cluster.
                         To set a key, run \'swift post -m
                         "Temp-URL-Key:b3968d0207b54ece87cccc06515a89d4"\'
+
+Optional arguments:
+  --absolute            Interpet the <seconds> positional argument as a Unix
+                        timestamp rather than a number of seconds in the
+                        future.
 '''.strip('\n')
 
 
 def st_tempurl(parser, args, thread_manager):
+    parser.add_option(
+        '--absolute', action='store_true',
+        dest='absolute_expiry', default=False,
+        help=("If present, seconds argument will be interpreted as a Unix "
+              "timestamp representing when the tempURL should expire, rather "
+              "than an offset from the current time")
+    )
     (options, args) = parse_args(parser, args)
     args = args[1:]
     if len(args) < 4:
@@ -1013,7 +1058,8 @@ def st_tempurl(parser, args, thread_manager):
         thread_manager.print_msg('WARNING: Non default HTTP method %s for '
                                  'tempurl specified, possibly an error' %
                                  method.upper())
-    url = generate_temp_url(path, seconds, key, method)
+    url = generate_temp_url(path, seconds, key, method,
+                            absolute=options.absolute_expiry)
     thread_manager.print_msg(url)
 
 
@@ -1110,7 +1156,7 @@ def main(arguments=None):
     version = client_version
     parser = OptionParser(version='python-swiftclient %s' % version,
                           usage='''
-usage: %%prog [--version] [--help] [--os-help] [--snet] [--verbose]
+usage: %prog [--version] [--help] [--os-help] [--snet] [--verbose]
              [--debug] [--info] [--quiet] [--auth <auth_url>]
              [--auth-version <auth_version>] [--user <username>]
              [--key <api_key>] [--retries <num_retries>]
@@ -1130,7 +1176,7 @@ usage: %%prog [--version] [--help] [--os-help] [--snet] [--verbose]
              [--os-endpoint-type <endpoint-type>]
              [--os-cacert <ca-certificate>] [--insecure]
              [--no-ssl-compression]
-             <subcommand> [--help]
+             <subcommand> [--help] [<subcommand options>]
 
 Command-line interface to the OpenStack Swift API.
 
@@ -1150,29 +1196,29 @@ Positional arguments:
     auth                 Display auth related environment variables.
 
 Examples:
-  %%prog download --help
+  %prog download --help
 
-  %%prog -A https://auth.api.rackspacecloud.com/v1.0 -U user -K api_key stat -v
+  %prog -A https://auth.api.rackspacecloud.com/v1.0 -U user -K api_key stat -v
 
-  %%prog --os-auth-url https://api.example.com/v2.0 --os-tenant-name tenant \\
+  %prog --os-auth-url https://api.example.com/v2.0 --os-tenant-name tenant \\
       --os-username user --os-password password list
 
-  %%prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
+  %prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
       --os-project-name project1 --os-project-domain-name domain1 \\
       --os-username user --os-user-domain-name domain1 \\
       --os-password password list
 
-  %%prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
+  %prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
       --os-project-id 0123456789abcdef0123456789abcdef \\
       --os-user-id abcdef0123456789abcdef0123456789 \\
       --os-password password list
 
-  %%prog --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
+  %prog --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
       --os-storage-url https://10.1.5.2:8080/v1/AUTH_ced809b6a4baea7aeab61a \\
       list
 
-  %%prog list --lh
-'''.strip('\n') % globals())
+  %prog list --lh
+'''.strip('\n'))
     parser.add_option('--os-help', action='store_true', dest='os_help',
                       help='Show OpenStack authentication options.')
     parser.add_option('--os_help', action='store_true', help=SUPPRESS_HELP)
@@ -1373,6 +1419,7 @@ Examples:
         logging.getLogger("swiftclient")
         if options.debug:
             logging.basicConfig(level=logging.DEBUG)
+            logging.getLogger('iso8601').setLevel(logging.WARNING)
         elif options.info:
             logging.basicConfig(level=logging.INFO)
 
