@@ -17,9 +17,9 @@ import hashlib
 import hmac
 import json
 import logging
-import time
-
 import six
+import time
+import traceback
 
 TRUE_VALUES = set(('true', '1', 'yes', 'on', 't', 'y'))
 EMPTY_ETAG = 'd41d8cd98f00b204e9800998ecf8427e'
@@ -35,49 +35,45 @@ def config_true_value(value):
         (isinstance(value, six.string_types) and value.lower() in TRUE_VALUES)
 
 
-def prt_bytes(bytes, human_flag):
+def prt_bytes(num_bytes, human_flag):
     """
     convert a number > 1024 to printable format, either in 4 char -h format as
     with ls -lh or return as 12 char right justified string
     """
 
-    if human_flag:
-        suffix = ''
-        mods = list('KMGTPEZY')
-        temp = float(bytes)
-        if temp > 0:
-            while temp > 1023:
-                try:
-                    suffix = mods.pop(0)
-                except IndexError:
-                    break
-                temp /= 1024.0
-            if suffix != '':
-                if temp >= 10:
-                    bytes = '%3d%s' % (temp, suffix)
-                else:
-                    bytes = '%.1f%s' % (temp, suffix)
-        if suffix == '':    # must be < 1024
-            bytes = '%4s' % bytes
+    if not human_flag:
+        return '%12s' % num_bytes
+
+    num = float(num_bytes)
+    suffixes = [None] + list('KMGTPEZY')
+    for suffix in suffixes[:-1]:
+        if num <= 1023:
+            break
+        num /= 1024.0
     else:
-        bytes = '%12s' % bytes
+        suffix = suffixes[-1]
 
-    return bytes
+    if not suffix:  # num_bytes must be < 1024
+        return '%4s' % num_bytes
+    elif num >= 10:
+        return '%3d%s' % (num, suffix)
+    else:
+        return '%.1f%s' % (num, suffix)
 
 
-def generate_temp_url(path, seconds, key, method):
-    """ Generates a temporary URL that gives unauthenticated access to the
+def generate_temp_url(path, seconds, key, method, absolute=False):
+    """Generates a temporary URL that gives unauthenticated access to the
     Swift object.
 
     :param path: The full path to the Swift object. Example:
-    /v1/AUTH_account/c/o.
+        /v1/AUTH_account/c/o.
     :param seconds: The amount of time in seconds the temporary URL will
-    be valid for.
-    :param key: The secret temporary URL key set on the Swift cluster.
-    To set a key, run 'swift post -m
-    "Temp-URL-Key:b3968d0207b54ece87cccc06515a89d4"'
-    :param method: A HTTP method, typically either GET or PUT, to allow for
-    this temporary URL.
+        be valid for.
+    :param key: The secret temporary URL key set on the Swift
+        cluster. To set a key, run 'swift post -m
+        "Temp-URL-Key: <substitute tempurl key here>"'
+    :param method: A HTTP method, typically either GET or PUT, to allow
+        for this temporary URL.
     :raises: ValueError if seconds is not a positive integer
     :raises: TypeError if seconds is not an integer
     :return: the path portion of a temporary URL
@@ -85,7 +81,10 @@ def generate_temp_url(path, seconds, key, method):
     if seconds < 0:
         raise ValueError('seconds must be a positive integer')
     try:
-        expiration = int(time.time() + seconds)
+        if not absolute:
+            expiration = int(time.time() + seconds)
+        else:
+            expiration = int(seconds)
     except TypeError:
         raise TypeError('seconds must be an integer')
 
@@ -119,6 +118,20 @@ def parse_api_response(headers, body):
     return json.loads(body.decode(charset))
 
 
+def report_traceback():
+    """
+    Reports a timestamp and full traceback for a given exception.
+
+    :return: Full traceback and timestamp.
+    """
+    try:
+        formatted_lines = traceback.format_exc()
+        now = time.time()
+        return formatted_lines, now
+    except AttributeError:
+        return None, None
+
+
 class NoopMD5(object):
     def __init__(self, *a, **kw):
         pass
@@ -135,7 +148,7 @@ class ReadableToIterable(object):
     Wrap a filelike object and act as an iterator.
 
     It is recommended to use this class only on files opened in binary mode.
-    Due to the Unicode changes in python 3 files are now opened using an
+    Due to the Unicode changes in Python 3, files are now opened using an
     encoding not suitable for use with the md5 class and because of this
     hit the exception on every call to next. This could cause problems,
     especially with large files and small chunk sizes.
@@ -183,7 +196,7 @@ class LengthWrapper(object):
     """
     Wrap a filelike object with a maximum length.
 
-    Fix for https://github.com/kennethreitz/requests/issues/1648
+    Fix for https://github.com/kennethreitz/requests/issues/1648.
     It is recommended to use this class only on files opened in binary mode.
     """
     def __init__(self, readable, length, md5=False):
