@@ -124,7 +124,7 @@ def fake_http_connect(*code_iter, **kwargs):
         def getheaders(self):
             if self.headers:
                 return self.headers.items()
-            headers = {'content-length': len(self.body),
+            headers = {'content-length': str(len(self.body)),
                        'content-type': 'x-application/test',
                        'x-timestamp': self.timestamp,
                        'last-modified': self.timestamp,
@@ -132,7 +132,7 @@ def fake_http_connect(*code_iter, **kwargs):
                        'etag':
                        self.etag or '"%s"' % EMPTY_ETAG,
                        'x-works': 'yes',
-                       'x-account-container-count': 12345}
+                       'x-account-container-count': '12345'}
             if not self.timestamp:
                 del headers['x-timestamp']
             try:
@@ -263,9 +263,8 @@ class MockHttpTest(testtools.TestCase):
                         conn.resp.status = status
                     if auth_token:
                         headers = args[1]
-                        self.assertTrue('X-Auth-Token' in headers)
-                        actual_token = headers.get('X-Auth-Token')
-                        self.assertEqual(auth_token, actual_token)
+                        self.assertEqual(auth_token,
+                                         headers.get('X-Auth-Token'))
                     if query_string:
                         self.assertTrue(url.endswith('?' + query_string))
                     if url.endswith('invalid_cert') and not insecure:
@@ -325,7 +324,7 @@ class MockHttpTest(testtools.TestCase):
             self.orig_assertEqual(body, real_request['body'], err_msg)
 
         if len(expected) > 3:
-            headers = expected[3]
+            headers = CaseInsensitiveDict(expected[3])
             for key, value in headers.items():
                 real_request['key'] = key
                 real_request['expected_value'] = value
@@ -336,16 +335,30 @@ class MockHttpTest(testtools.TestCase):
                     'for %(method)s %(path)s %(headers)r' % real_request)
                 self.orig_assertEqual(value, real_request['value'],
                                       err_msg)
+            real_request['extra_headers'] = dict(
+                (key, value) for key, value in real_request['headers'].items()
+                if key not in headers)
+            if real_request['extra_headers']:
+                self.fail('Received unexpected headers for %(method)s '
+                          '%(path)s, got %(extra_headers)r' % real_request)
 
     def assertRequests(self, expected_requests):
         """
         Make sure some requests were made like you expected, provide a list of
         expected requests, typically in the form of [(method, path), ...]
+        or [(method, path, body, headers), ...]
         """
         real_requests = self.iter_request_log()
         for expected in expected_requests:
             real_request = next(real_requests)
             self.assert_request_equal(expected, real_request)
+        try:
+            real_request = next(real_requests)
+        except StopIteration:
+            pass
+        else:
+            self.fail('At least one extra request received: %r' %
+                      real_request)
 
     def assert_request(self, expected_request):
         """
@@ -503,7 +516,7 @@ class FakeKeystone(object):
         self.endpoint = endpoint
         self.token = token
 
-    class _Client():
+    class _Client(object):
         def __init__(self, endpoint, token, **kwargs):
             self.auth_token = token
             self.endpoint = endpoint
